
// scplay.q: play a SuperCollider synth via the OSC interface

////////////////////////////////////////////////////////////////////////////
//
// Poor man's monotimbral synth with 32 voices.
//
// Instructions:
//
// 0. Run the `compile' function to compile the synth definitions in the
//    synths.sc file. This only needs to be done once to get the binary
//    synth definition files in the synthdefs subdirectory, which can then
//    be loaded into the sound server. You'll also have to repeat this step
//    after making changes to the synths.sc file.
//
// 1. Start up a sound server somewhere and connect to it. E.g.: `sc_start'
//    both starts the local server and connects to it in one step. Make sure
//    you have `jackd' running before doing this.
//
// 2. Invoke `play NAME' where NAME is the name of a synth in the synthdefs
//    directory (created in step 0) you want to play, without the .scsyndef
//    suffix. E.g., use `play "add"' to play the additive synth defined in
//    synths.sc, `play "string"' to play the Karplus-Strong synth, etc.
//    Instead of `play NAME', you can also run `scope NAME' to display the
//    audio output in a GGI visual.
//
// 3. You can now play the synth via your MIDI keyboard, or connect other
//    MidiShare applications like the Q-Midi player to the "SuperCollider"
//    client via msconnect. There are also some global control values (GAIN,
//    PAN, REV) you can set before invoking `play'; see below for details.
//
// 4. Take a look at the synth definitions in synths.sc, fiddle around with
//    them, develop your own synths. Enjoy! :)
//
////////////////////////////////////////////////////////////////////////////

import midi, osc, sc;

// MIDI interface

def DRV = ifelse (pos "linux" sysinfo >= 0)
  (midi_client_ref "MidiShare/ALSA Bridge") 0,
  IN = DRV, OUT = DRV, PORT = 0,
  REF = midi_open "SuperCollider", CTRL = midi_open "SuperCollider - Ctrl",
  _ = do (flip midi_connect REF) [IN,CTRL] || midi_connect REF OUT ||
  midi_accept_type REF active_sense false ||
  // we don't want to receive any drum voices (MIDI channel #10)
  midi_accept_chan REF 9 false;

midiflush	= midi_flush REF;
midiavail	= midi_avail REF;
midiin		= (TIME,MSG) where (_,_,TIME,MSG) = midi_get REF;
midiout MSG	= midi_send REF PORT MSG;
midioffch CH	= midiout (ctrl_change CH 123 0);
midioff		= do midioffch [0..15];
midictrl MSG	= midi_send CTRL PORT MSG;
midiwait T	= midi_wait REF T;
miditime	= midi_time;

// Run this to (re)compile the synths.

compile		= sc_source "synths.sc";

// Set global control values for the synths here:

// - GAIN  = global gain (0 = silent, < 1 = attenuate, > 1 = amplify)
// - PAN   = pan (-1 = left .. 0 = center .. 1 = right)
// - REV   = reverb (>= 0, 0 = no reverb)
// - SCALE = Y scaling for the scope (= gain for the disk unit)

def GAIN = 1/2, PAN = 0, REV = 0, SCALE = 1;

// Convenience functions to invoke the OSC messages we need.

// load a synthdef into the server

d_load NAME	= sc_send (osc_message CMD_D_LOAD NAME);

// send a synthdef to the server

d_recv NAME	= sc_send (osc_message "/d_recv"
			   (bread (fopen NAME "rb")
			    (st_size (stat NAME))));

// create a new synth node (add at the end of the main group)

s_new NAME ID ARGS
		= sc_send (osc_message CMD_S_NEW (NAME,ID,1,0|ARGS));

// free a synth node

n_free ID	= sc_send (osc_message CMD_N_FREE ID);

// set control parameters

n_set ID ARGS	= sc_send (osc_message CMD_N_SET (ID|ARGS));

// allocate a sample buffer

b_alloc BUF NBUF NCHAN
		= sc_send (osc_message CMD_B_ALLOC (BUF,NBUF,NCHAN));

// associate buffer with an output file/pipe

b_write BUF NAME
		= sc_send (osc_message CMD_B_WRITE
			   (BUF,NAME,"raw","int16",-1,0,1));

// show current voice allocation

show P Q	= printf "%d: %s\n" (miditime,join " " (map str (list P)));

// loop P Q EV: process incoming MIDI messages

// - P  = dictionary of allocated (playing) synths I, indexed by notes N
// - Q  = queue of unallocated (quiet) synths waiting to be (re)used
// - EV = next MIDI event to be processed

// Voices are allocated in a round-robin fashion (taken from the beginning of
// the queue when needed to play a note, placed back to the end of the queue
// when the note is turned off). By these means we (hopefully) give the queued
// voice enough time to finish its envelope before it is allocated again.

//loop _ _ MSG	= printf "msg:  %s\n" (str MSG) || fail;

loop _ _ (_,stop)
		= halt;

// note-offs: set gate to 0, deallocate the voice and put it at the end of the
// queue

loop P Q (_,note_on _ N 0)
		= n_set I ("gate",0) || show P Q || loop P Q midiin
		    where I:Int = P!N, P = delete P N, Q = append Q I;
		= loop P Q midiin otherwise;

loop P Q (T,note_off CH N _)
		= loop P Q (T,note_on CH N 0);

// note-ons: turn note off if already sounding, then allocate a new voice from
// the queue and set its gate to 1

loop P Q (T,note_on CH N V)
		= n_set I ("gate",0) || loop P Q (T,note_on CH N V)
		    where I:Int = P!N, P = delete P N, Q = append Q I;
		= n_set I ("freq",FREQ,"gain",V/127,"gate",1) ||
		  show P Q || loop P Q midiin
		    where [I|Q] = Q, P = insert P (N,I),
		      FREQ = 440*2^((N-69)/12);

// translate MidiShare note messages, schedule note-offs for later

loop P Q (T,note CH N V D)
		= midictrl (T+D,note_on CH N 0) ||
		  loop P Q (T,note_on CH N V);

// ignore all other messages

loop P Q  _	= loop P Q midiin otherwise;

// perform realtime scheduling (requires privileges)

private maxprio;

def POL = SCHED_RR, PRIO = maxprio;

// check for the maximum available priority <= 10

testprio PRIO	= false if PRIO > 10;
		= setsched this_thread 0 0 || true
		    where () = setsched this_thread POL PRIO;
		= false otherwise;

maxprio		= until (neg testprio) (+1) 1 - 1;

special realtime X;

realtime X	= setsched TH 0 0 || Y
  		    where TH = this_thread, Y = setsched TH POL PRIO || X;
		  
// Invoke this with the filename of the binary synthdef file (.scsyndef is
// added automatically), e.g.: play "sub".

// Note that synth nodes are numbered 11..42 instead of 1..32, to prevent
// clashes with preallocated nodes in some SC3 versions.

play NAME	= // send the synthdefs to the server, wait for completion
		  d_recv ("synthdefs/"++NAME++".scsyndef") || sc_recv ||
		  d_recv "synthdefs/reverb.scsyndef" || sc_recv ||
		  d_recv "synthdefs/amp.scsyndef" || sc_recv ||
		  // create synth nodes for 32 voices
		  do (flip (s_new NAME) ("pan",PAN)) [11..42] ||
		  // create the effect and postprocessing units
		  s_new "reverb" 90 ("rev", REV) ||
		  s_new "amp" 98 ("gain", GAIN) ||
		  // empty the MIDI input buffer
		  midiflush ||
		  // run the MIDI input loop
		  realtime (catch (cst ()) (loop emptydict [11..42] midiin)) ||
		  // free allocated resources
		  do n_free [11..42] || do n_free [90,98] || sleep 0.5;

// scope_vis: simple display of audio output in a GGI visual

import audio, ggi, draw_wave;

// Create a pipe which will be used to send the audio output from the sound
// server to the scope_loop function. To make this work, the sound server must
// run on the same machine!

def FIFO = tmpnam, () = unlink FIFO || mkfifo FIFO 0666;

// buffer size for audio output

def BUFSZ = 2048;

// colors and dimensions of the GGI visual

def BACK = (0,0,0), FORE = (0xffff,0xffff,0xffff), WAVCOL = (0,0xffff,0),
  WD = 320, HT = 240, WH = (HT-30) div 2, LT = 10, LC = LT+WH div 2,
  RT = LT+10+WH, RC = RT+WH div 2;

// display a screenful of the wave in the visual; assumes that WAV is a byte
// string of 16 bit integer stereo samples

show_wave VIS WAV
		= ggi_set_foreground VIS BACK ||
		  ggi_draw_box VIS (0,LT) (WD,WH) ||
		  ggi_draw_box VIS (0,RT) (WD,WH) ||
		  ggi_set_foreground VIS WAVCOL ||
		  ggi_draw_hline VIS (0,LC) WD ||
		  ggi_draw_hline VIS (0,RC) WD ||
		  draw_wave VIS (0,LT) (WD,WH) 0 2 PA_INT16 WAV ||
		  ggi_puts VIS (5,LT) "L" ||
		  draw_wave VIS (0,RT) (WD,WH) 1 2 PA_INT16 WAV ||
		  ggi_puts VIS (5,RT) "R" ||
		  ggi_flush VIS;

// read a screenful of samples from the pipe

read_wave F N W	= W if N<=0;
		= read_wave F (N-#WAV) (W++WAV)
		    where WAV:ByteStr = bread F N if N<=BUFSZ;
		= read_wave F (N-#WAV) (W++WAV)
		    where WAV:ByteStr = bread F BUFSZ;

// repeat reading and displaying the wave until we get killed or the pipe is
// closed

scope_loop VIS F
		= show_wave VIS WAV || scope_loop VIS F if #WAV>0
		    where WAV:ByteStr =
		      read_wave F (4*BUFSZ) (bytestr []);
		= () otherwise;

// initialize the visual and kick off the display loop

scope_vis	= scope_loop VIS F
		    where F:File = fopen FIFO "rb",
		      VIS:GGIVisual = ggi_open "x:-keepcursor",
		      _ = ggi_set_mode VIS (sprintf "%dx%d" (WD,HT)) ||
		      ggi_set_flags VIS GGI_FLAG_ASYNC,
		      _ = ggi_set_background VIS BACK ||
		      ggi_set_foreground VIS FORE;

// Entry point for the scope. Same as `play', but also arrange for audio
// output from the server to go to the pipe and display the scope. NOTE: This
// will only work with a local sound server.

scope NAME	= d_recv ("synthdefs/"++NAME++".scsyndef") || sc_recv ||
		  d_recv "synthdefs/reverb.scsyndef" || sc_recv ||
		  d_recv "synthdefs/amp.scsyndef" || sc_recv ||
		  // add the disk output unit
		  d_recv "synthdefs/disk.scsyndef" || sc_recv ||
		  // allocate a buffer and associate it with output to a pipe
		  b_alloc 0 BUFSZ 2 || sc_recv ||
		  b_write 0 FIFO || sc_recv ||
		  // create synth nodes (including the disk unit)
		  do (flip (s_new NAME) ("pan",PAN)) [11..42] ||
		  s_new "reverb" 90 ("rev", REV) ||
		  s_new "amp" 98 ("gain", GAIN) ||
		  s_new "disk" 99 ("gain", SCALE) ||
		  midiflush ||
		  realtime (catch (cst ()) (loop emptydict [11..42] midiin)) ||
		  // kill the scope display thread and release resources
		  cancel H ||
		  do n_free [11..42] || do n_free [90,98,99] || sleep 0.5
		    where // invoke the scope display as a background task
		      H = thread scope_vis;
