#!/usr/bin/q --
#! -c main

// scgui.q: play a SuperCollider synth via the OSC interface
// This is an enhanced GUI version of the the scplay.q example.

// set this to your preferred text editor

def EDITOR = "xemacs";

// perform realtime scheduling (requires privileges) ////////////////////////

private maxprio;

def POL = SCHED_RR, PRIO = maxprio;

// check for the maximum available priority <= 10

testprio PRIO	= false if PRIO > 10;
		= setsched this_thread 0 0 || true
		    where () = setsched this_thread POL PRIO;
		= false otherwise;

maxprio		= until (neg testprio) (+1) 1 - 1;

special realtime X;

realtime_on	= setsched this_thread POL PRIO;
realtime_off	= setsched this_thread 0 0;

realtime X	= realtime_off || Y where Y = realtime_on || X;

// MIDI interface ///////////////////////////////////////////////////////////

import midi;

client NAME 1	= NAME;
client NAME N	= sprintf "%s #%d" (NAME,N);

clientno NAME N	= N if not isint (midi_client_ref NAME)
		    where NAME = client NAME N;
		= clientno NAME (N+1) otherwise;

def DRV = ifelse (pos "linux" sysinfo >= 0)
  (midi_client_ref "MidiShare/ALSA Bridge") 0,
  IN = DRV, OUT = DRV, PORT = 0,
  CLIENTNO = clientno "SuperCollider" 1,
  CLIENT = client "SuperCollider" CLIENTNO,
  REF = midi_open CLIENT, CTRL = midi_open (CLIENT++" - Ctrl"),
  _ = do (flip midi_connect REF) [IN,CTRL] || midi_connect REF OUT ||
  midi_accept_type REF active_sense false ||
  midi_accept_chan REF 9 false;

midiflush	= midi_flush REF;
midiavail	= midi_avail REF;
midiin		= (TIME,MSG) where (_,_,TIME,MSG) = midi_get REF;
midiout MSG	= midi_send REF PORT MSG;
midioffch CH	= midiout (ctrl_change CH 123 0);
midioff		= do midioffch [0..15];
midictrl MSG	= midi_send CTRL PORT MSG;
midiwait T	= midi_wait REF T;
miditime	= midi_time;

// OSC interface ////////////////////////////////////////////////////////////

import osc, sc;

// global control values

// - NVOICES = number of voices

// - GAIN  = global gain (0 = silent, < 1 = attenuate, > 1 = amplify)
// - PAN   = pan (-1 = left .. 0 = center .. 1 = right)
// - REV   = reverb (>= 0, 0 = no reverb)
// - REC   = recording level

def NVOICES = ref 16, GAIN = ref 0.5, PAN = ref 0, REV = ref 0, REC = ref 1;

// convenience functions to invoke the OSC messages we need

// load a synthdef into the server

d_load NAME	= sc_send (osc_message CMD_D_LOAD NAME);

// send a synthdef to the server

d_recv NAME	= sc_send (osc_message "/d_recv"
			   (bread (fopen NAME "rb")
			    (st_size (stat NAME))));

// create a new synth node (add at the end of the main group)

s_new NAME ID ARGS
		= sc_send (osc_message CMD_S_NEW (NAME,ID,1,0|ARGS));

// free a synth node

n_free ID	= sc_send (osc_message CMD_N_FREE ID);

// set control parameters

n_set ID ARGS	= sc_send (osc_message CMD_N_SET (ID|ARGS));

// allocate a sample buffer

b_alloc BUF NBUF NCHAN
		= sc_send (osc_message CMD_B_ALLOC (BUF,NBUF,NCHAN));

// free buffer

b_free BUF	= sc_send (osc_message CMD_B_FREE BUF);

// associate buffer with an output file/pipe

b_write BUF NAME FORMAT
		= sc_send (osc_message CMD_B_WRITE
			   (BUF,NAME,FORMAT,"int16",-1,0,1));

// receive with timeout to prevent lockups

set_nonblock FD	= fcntl FD F_SETFL (FLAGS or O_NONBLOCK)
		    where FLAGS = fcntl FD F_GETFL ();

sc_recvnb T	= osc_decode B
		    where B:ByteStr = set_nonblock SC || recv SC 0 500000;
		= sc_recvnb T if time < T;

// FIXME: timeout value should be configurable
def TIMEOUT = 5;

my_recv		= sc_recvnb (time+TIMEOUT);

// global shared data ///////////////////////////////////////////////////////

// list of available synths

basename EXT NAME
		= substr NAME 0 (#NAME-#EXT)
		    where NAME = last (split "/" NAME);

chksynth NAME	= not any (=NAME) ["default","disk","amp","chorus","reverb"];

synths		= filter chksynth
		  (map (basename ".scsyndef") (glob "synthdefs/*.scsyndef"));

def SYNTHS = synths,
  SYNTH = ref (ifelse (#SYNTHS>0) (hd SYNTHS) ());

// message semaphore, misc state variables

def MSGS = sem, PLAYING = ref false, RECORDING = ref false, TIME = ref (),
  LASTV = ref (emptydict,[1001..get NVOICES+1000]),
  LASTBUF = ref ();

// pipe for monitoring audio output

def FIFO = tmpnam, _ = unlink FIFO || mkfifo FIFO 0666, BUFSZ = ref 4096;

private open_rec, close_rec, write_rec WAV;

// MIDI event loop //////////////////////////////////////////////////////////

private stopped EV, loop P Q EV, show P Q;
// - P  = dictionary of allocated (playing) synths I, indexed by notes N
// - Q  = queue of unallocated (quiet) synths waiting to be (re)used
// - EV = next MIDI event to be processed

// this loop is executed while the synth is stopped

stopped (_,start)
		= realtime_off ||
		  d_recv ("synthdefs/"++NAME++".scsyndef") || my_recv ||
		  d_recv "synthdefs/reverb.scsyndef" || my_recv ||
		  d_recv "synthdefs/amp.scsyndef" || my_recv ||
		  // add the disk output unit
		  d_recv "synthdefs/disk.scsyndef" || my_recv ||
		  ifelse (get RECORDING) open_rec () ||
		  // allocate a buffer and associate it with output to a pipe
		  b_alloc 0 (get BUFSZ) 2 || my_recv ||
		  b_write 0 FIFO "raw" || my_recv ||
		  // create synth nodes (including the disk unit)
		  do (flip (s_new NAME) ("pan",get PAN))
		  [1001..get NVOICES+1000] ||
		  s_new "reverb" 2090 ("rev", get REV) ||
		  s_new "amp" 2098 ("gain", get GAIN) ||
		  s_new "disk" 2099 ("gain", get REC) ||
		  put PLAYING true ||
		  ifelse (eq () (get TIME)) (put TIME miditime) () ||
		  post MSGS (text "Playing") ||
		  show emptydict [1001..get NVOICES+1000] ||
		  realtime_on ||
		  loop emptydict [1001..get NVOICES+1000] midiin
		    if not null NAME
		    where NAME = get SYNTH;

stopped (_,text "R")
		= put RECORDING (not get RECORDING) || post MSGS (text "R") ||
		  stopped midiin;

stopped _	= stopped midiin;

// the following loop is executed while the synth is running

// Voices are allocated in a round-robin fashion (taken from the beginning of
// the queue when needed to play a note, placed back to the end of the queue
// when the note is turned off). By these means we (hopefully) give the queued
// voice enough time to finish its envelope before it is allocated again.

// Note that the voices are actually numbered 1001..NVOICES+1000, to prevent
// clashes with preallocated nodes in some SC3 versions.

// note-offs: set gate to 0, deallocate the voice and put it at the end of the
// queue

loop P Q (_,note_on _ N 0)
		= n_set I ("gate",0) || show P Q || loop P Q midiin
		    where I:Int = P!N, P = delete P N, Q = append Q I;
		= loop P Q midiin otherwise;

loop P Q (T,note_off CH N _)
		= loop P Q (T,note_on CH N 0);

// note-ons: turn note off if already sounding, then allocate a new voice from
// the queue and set its gate to 1

loop P Q (T,note_on CH N V)
		= n_set I ("gate",0) || loop P Q (T,note_on CH N V)
		    where I:Int = P!N, P = delete P N, Q = append Q I;
		= n_set I ("freq",FREQ,"gain",V/127,"gate",1) ||
		  show P Q || loop P Q midiin
		    where [I|Q] = Q, P = insert P (N,I),
		      FREQ = 440*2^((N-69)/12);

// translate MidiShare note messages, schedule note-offs for later

loop P Q (T,note CH N V D)
		= midictrl (T+D,note_on CH N 0) ||
		  loop P Q (T,note_on CH N V);

// recording

loop P Q (_,text "R")
		= put RECORDING (not get RECORDING) || post MSGS (text "R") ||
		  ifelse (get RECORDING) open_rec close_rec ||
		  loop P Q midiin;

// stop/restart the synth

loop P Q (_,stop)
		= realtime_off ||
		  ifelse (get RECORDING)
		  (close_rec || put RECORDING false || post MSGS (text "R"))
		  () ||
		  do n_free (vals P++Q) || do n_free [2090,2098,2099] ||
		  b_free 0 ||
		  put PLAYING false || put TIME () ||
		  post MSGS (text "Stopped") ||
		  realtime_on ||
		  stopped midiin;

loop P Q (_,continue)
		= realtime_off ||
		  ifelse (get RECORDING)
		  (close_rec || put RECORDING false || post MSGS (text "R"))
		  () ||
		  do n_free (vals P++Q) || do n_free [2090,2098,2099] ||
		  b_free 0 ||
		  put PLAYING false ||
		  post MSGS (text "Stopped") ||
		  realtime_on ||
		  stopped (miditime,start);

// ignore all other messages

loop P Q  _	= loop P Q midiin otherwise;

// the main entry point to the play loop

play		= realtime (midiflush || stopped midiin);

// the GUI //////////////////////////////////////////////////////////////////

import audio, sndfile, ggi, draw_wave, tk;

target		= "directx" if pos "mingw" sysinfo >= 0;
		= "x" otherwise;

// initialize the application

prefix NAME	= substr ANAME 0 (#ANAME-2)
		    where ANAME:String = which NAME;
		= NAME otherwise;

update_synths	= tk (sprintf "$widget(SynthComboBox) configure -values {%s}"
		      (join " " SYNTHS))
		    where SYNTHS = synths;

def WIN = tk "wm withdraw .; set argc 0; set argv {}" ||
  tk (sprintf "source %s" (str (prefix "scgui"++".tcl"))) ||
  tk "proc exit { {returnCode 0} } { q scgui::quit_cb }" ||
  tk (sprintf "wm title $widget(Toplevel1) {%s}" CLIENT) ||
  tk "set filename {}; set voices 16; set server Boot; set status Ready" ||
  do (uncurry tk_set)
  (zip ["voices","bufsz","vol","pan","rev","rec"]
   (map (str.get) [NVOICES,BUFSZ,GAIN,PAN,REV,REC])) ||
  tk "$widget(ChanComboBox) setvalue first" ||
  tk "winfo id $widget(Wave)";

// wave display and recording functions

def BACK = (0xffff,0xffff,0xffff), FORE = (0xffff,0x5959,0x9696),
  WAVCOL = (0x9696,0xabab,0xffff),
  VIS = ggi_open (sprintf "%s:-keepcursor -inwin=%s" (target,WIN)),
  _ = ggi_set_mode VIS "" || ggi_set_flags VIS GGI_FLAG_ASYNC,
  _ = ggi_set_background VIS BACK || ggi_set_foreground VIS FORE ||
      ggi_clear VIS || ggi_flush VIS,
  (WD,HT) = sscanf (ggi_get_mode VIS) "%dx%d",
  WH = (HT-40) div 2, LT = 15, LC = LT+WH div 2,
  RT = LT+20+WH, RC = RT+WH div 2;

show_voice DX I	= ggi_draw_box VIS ((I-1001)*8+DX,HT-6) (4,4);

show P Q	= ggi_set_foreground VIS BACK ||
		  ggi_draw_box VIS (0,HT-8) (WD,8) ||
		  ggi_set_foreground VIS FORE ||
		  do (show_voice DX) P ||
		  ggi_set_foreground VIS WAVCOL ||
		  do (show_voice DX) Q ||
		  ggi_flush VIS
		    where _ = put LASTV (P,Q),
		      P = vals P, DX = (WD-(#P+#Q)*8) div 2;

show_time	= ggi_puts VIS (WD-200,0)
		  (sprintf "%02d:%02d" (MIN,SEC))
		    where T0:Int = get TIME, T1 = miditime, DT = T1-T0,
		      DT = ifelse (DT<0) (DT+0x100000000) DT,
		      DT = DT div 1000, MIN = DT div 60, SEC = DT mod 60;
		= ggi_puts VIS (WD-200,0) "      " otherwise;
		  
show_cpu AVG PEAK
		= ggi_set_foreground VIS WAVCOL ||
		  show_time ||
		  ggi_puts VIS (WD-100,0)
		  (sprintf "CPU %4.1f%%/%4.1f%%" (AVG,PEAK)) ||
		  ggi_flush VIS;

// display a screenful of the wave in the visual; assumes that WAV is a byte
// string of 16 bit integer stereo samples

show_wave WAV	= ggi_set_foreground VIS BACK ||
		  ggi_draw_box VIS (0,LT) (WD,WH) ||
		  ggi_draw_box VIS (0,RT) (WD,WH) ||
		  ggi_set_foreground VIS WAVCOL ||
		  ggi_draw_hline VIS (0,LC) WD ||
		  ggi_draw_hline VIS (0,RC) WD ||
		  draw_wave VIS (0,LT) (WD,WH) 0 2 PA_INT16 WAV ||
		  ggi_puts VIS (5,LT) "L" ||
		  draw_wave VIS (0,RT) (WD,WH) 1 2 PA_INT16 WAV ||
		  ggi_puts VIS (5,RT) "R" ||
		  ggi_flush VIS;

// generate filenames for recordings

newfile N	= NAME if not isfile (fopen NAME "r")
		    where NAME = sprintf "scgui%03d.wav" N;
		= newfile (N+1) otherwise;

// recording functions

def RECFILE = ref ();

open_rec	= close_rec || open_rec if is_snd_file (get RECFILE);
		= put RECFILE F
		    where F:SndFile = sf_open (newfile 0) SFM_WRITE
		      (44100, 2, SF_FORMAT_WAV or SF_FORMAT_PCM_16);

close_rec	= put RECFILE () || sf_close F where F:SndFile = get RECFILE;

write_rec WAV	= sf_write_short F WAV where F:SndFile = get RECFILE;

// read a screenful of samples from the pipe

chk B		= sleep 0.05 || B if #B=0;
		= B;

read_wave F N BUF
		= BUF if N<=0;
		= read_wave F (N-#WAV) (BUF++WAV)
		    where WAV:ByteStr = chk (bread F N) if N<=get BUFSZ;
		= read_wave F (N-#WAV) (BUF++WAV)
		    where WAV:ByteStr = chk (bread F (get BUFSZ));

// repeat reading and displaying/recording the wave until we get killed or the
// pipe is closed

scope_loop F	= show_wave WAV || put LASTBUF WAV ||
		  ifelse (get RECORDING) (write_rec WAV) () ||
		  scope_loop F if #WAV>0
		    where WAV:ByteStr =
		      read_wave F (4*get BUFSZ) (bytestr []);
		= () otherwise;

// scope display/recording task

scope		= scope_loop F
		    where F:File = fopen FIFO "rb";

// inferior processes ///////////////////////////////////////////////////////

// run sclang and scsynth in pseudo terminals so we can capture their output
// and display it in a text window

termattr FD	= tcsetattr FD TCSAFLUSH ATTR
		    where ATTR = tcgetattr FD,
		      (IF,OF,CF,LF,IS,OS,CC) = ATTR,
		      LF = LF and not (ECHO or ECHOE or ECHOK or ECHONL),
		      OF = OF and ONLRET,
		      ATTR = (IF,OF,CF,LF,IS,OS,CC);

pty CMD		= (PID,MASTER) if PID > 0
		    where (PID,MASTER) = forkpty;
		= termattr 0 || exec "/bin/sh" ["/bin/sh", "-c", CMD]
		    otherwise;

reader FD	= post MSGS S || reader FD
		    where S:String = bstr (bread FD 1024);
		= () otherwise;

sclang		= (H,PID,MASTER)
		    where (PID,MASTER) = pty "sclang",
		      H = thread (reader MASTER);

scsynth		= (H,PID,MASTER)
		    where (PID,MASTER) =
		      pty (sprintf "scsynth -u %d" (57110+CLIENTNO-1)),
		      H = thread (reader MASTER);

fini (H,PID,MASTER)
		= close MASTER || waitpid PID WNOHANG || cancel H;
fini _		= ();

fini (LANG,SERVER,PLAYER)
		= fini LANG || fini SERVER || cancel PLAYER;

// status updates ///////////////////////////////////////////////////////////

print ()	= show_wave (get LASTBUF) || show P Q
		    where (P,Q) = get LASTV
		    if not get PLAYING;
		= ();
print (text "R")
		= tk "set recstat REC" if get RECORDING;
		= tk "set recstat {}" otherwise;
print (text S:String)
		= tk (sprintf "set status {%s}" S);
print S:String	= tk_set "line" S ||
		  tk "$widget(Output) insert end $line" ||
		  tk "$widget(Output) see end";

private check_status MSG;

get_status SERVER
		= check_status SERVER
		  (sc_send (osc_message "/status" ()) || my_recv);

check_status SERVER MSG
		= MSG where osc_message "status.reply" _ = MSG;
		= get_status SERVER if is_osc_packet MSG;
		= post MSGS
		  "Hmm, looks like scsynth hangs, better restart...\n" ||
		  kill SIGTERM PID || MSG where (_,PID,_) = SERVER
		    otherwise;

update_status ()
		= do (tk.sprintf "$widget(%sButton) configure -state disabled")
		  ["Play","Stop","Record"] ||
		  do (flip tk_set "") ["defs","groups","synths","ugens"] ||
		  show_cpu 0.0 0.0 || ();
update_status SERVER
		= do (tk.sprintf "$widget(%sButton) configure -state normal")
		  ["Play","Stop","Record"] ||
		  tk_set "defs" (str DEFS) ||
		  tk_set "groups" (str GROUPS) ||
		  tk_set "synths" (str SYNTHS) ||
		  tk_set "ugens" (str UGENS) ||
		  show_cpu AVG PEAK || ()
		    where
		      osc_message _ (_,UGENS,SYNTHS,GROUPS,DEFS,AVG,PEAK) =
		      get_status SERVER
		    if active H
		    where (H|_) = SERVER;

clear_synth	= put SYNTH ();

update_synth	= put SYNTH NAME
		    if (isfile (fopen ("synthdefs/"++NAME++".scsyndef") "r"))
		    where NAME:String = tk_get "synth";
		= clear_synth otherwise;

// the main program /////////////////////////////////////////////////////////

exception X	= fprintf ERROR "Exception: %s\n" (str X) || unlink FIFO ||
		  exit 1;

main_loop (LANG,SERVER,PLAYER)
		= tk "set status {Error: sclang process died unexpectedly}" ||
		  fini LANG ||
		  main_loop ((),SERVER,PLAYER)
		    if not active H
		    where (H|_) = LANG;
		= midictrl stop || sleep 0.1 ||
		  post MSGS
		  (text "Error: scsynth process died unexpectedly") ||
		  tk "set server Boot" || clear_synth || fini SERVER ||
		  main_loop (LANG,(),PLAYER)
		    if not active H
		    where (H|_) = SERVER;

main_loop STATE	= fini STATE || exit 0 if not tk_ready;
		= main_loop (tk_read STATE) if tk_check;
		= print (get MSGS) || main_loop STATE if #MSGS > 0;
		= update_status (STATE!1) || sleep 0.1 || main_loop STATE
		    otherwise;

main		= main_loop INIT || cancel H || unlink FIFO || exit 0
		    where INIT = show_wave () ||
		      show emptydict [1001..get NVOICES+1000] ||
		      ((),(),thread play),
		      H = thread scope,
		      NAME = ifelse (#ARGS > 1) (ARGS!1) "",
		      INIT = open INIT NAME,
		      _ = sleep 0.2 || update_synths ||
		      tk "$widget(SynthComboBox) setvalue first";

// callbacks ////////////////////////////////////////////////////////////////

quit_cb STATE	= fini STATE || exit 0;

expose_cb C STATE
		= show_wave (get LASTBUF) || show P Q || STATE
		    where (P,Q) = get LASTV
		    if not get PLAYING and then (C>0);
		= STATE otherwise;

wait_boot	= () if not tk_ready or else (#MSGS>0);
		= sleep 0.1 || wait_boot otherwise;

boot_cb (LANG,(),PLAYER)
		= wait_boot || sleep 0.2 ||
		  sc_connect ("localhost", 57110+CLIENTNO-1) ||
		  tk "set server Quit; set status {Server started}" ||
		  update_synth || (LANG,SERVER,PLAYER)
		    where SERVER:Tuple = scsynth;
		= tk "set status {Error starting server}" || (LANG,(),PLAYER)
		    otherwise;
boot_cb (LANG,SERVER,PLAYER)
		= midictrl stop ||
		  sc_send (osc_message "/quit" ()) || sleep 0.1 ||
		  fini SERVER || clear_synth ||
		  tk "set server Boot" || post MSGS (text "Server stopped") ||
		  (LANG,(),PLAYER);

open_dg		= tk "tk_getOpenFile -title \"Open...\" \
-parent $widget(Toplevel1) \
-defaultextension \".sc\" \
-filetypes {{{SCLang Files} {.sc}} {{All Files} *}}";

open STATE NAME	= ifelse (PWD<>PATH)
		  (tk_set "status" (sprintf "New working dir: %s" PATH)) () ||
		  tk_set "filename" NAME || (LANG,SERVER,PLAYER)
		    where PWD = getcwd, PATH = split "/" NAME,
		      PATH = ifelse (null PATH) [""] PATH,
		      NAME = last PATH, PATH = join "/" (init PATH),
		      PATH = ifelse (null PATH) PWD PATH,
		      () = chdir PATH,
		      //_ = mkdir "synthdefs" 0777,
		      (LANG,SERVER,PLAYER) = STATE,
		      LANG = ifelse ((PWD<>PATH) or else null LANG)
		      (fini LANG || sclang) LANG;

open_cb STATE	= post MSGS () || open STATE NAME
		    where NAME:String = open_dg;
		= post MSGS () || STATE otherwise;

edit_cb STATE	= system (sprintf "%s %s &" (EDITOR,str NAME)) ||
		  STATE
		    if not null NAME
		    where NAME:String = tk_get "filename";
		= tk "set status {No source file}" || STATE
		    otherwise;

compile_cb (LANG,SERVER,PLAYER)
		= mkdir "synthdefs" 0777 ||
		  bwrite MASTER (bytestr (SOURCE++"\f\n")) ||
		  tk "set status {Compiling source}" ||
		  (LANG,SERVER,PLAYER)
		    where F:File = fopen (tk_get "filename") "r",
		      SOURCE:String = fget F,
		      (_,_,MASTER) = LANG;
		= tk "set status {Error compiling source}" ||
		  (LANG,SERVER,PLAYER)
		    otherwise;

start_cb STATE	= midictrl start || STATE;

stop_cb STATE	= midictrl stop || STATE;

record_cb STATE	= midictrl (text "R") || STATE;

synth_cb STATE	= update_synth || midictrl continue || STATE
		    if not null (STATE!1);
		= clear_synth || STATE otherwise;

synth_post_cb STATE
		= // expose when the synth listbox is unmapped
		  tk "bind $widget(SynthComboBox).shell.listb <Unmap> \
{q scgui::expose_cb 1}" ||
		  update_synths || STATE;

chan_cb STATE	= do (flip (midi_accept_chan REF) true) [0..15] ||
		  midi_accept_chan REF 9 false || midictrl continue ||
		  STATE
		    if tk_get "chan" = "All";
		= do (flip (midi_accept_chan REF) false) [0..15] ||
		  midi_accept_chan REF (CHAN-1) true || midictrl continue ||
		  STATE
		    where CHAN:Int = val (tk_get "chan");
		= STATE otherwise;

chan_post_cb STATE
		= // expose when the channel listbox is unmapped
		  tk "bind $widget(ChanComboBox).shell.listb <Unmap> \
{q scgui::expose_cb 1}" || STATE;

voices_cb STATE	= put NVOICES N || midictrl continue || STATE
		    where N:Int = val (tk_get "voices");

pan_cb STATE	= put PAN VAL ||
		  ifelse (get PLAYING)
		  (do (flip n_set ("pan",VAL)) [1001..get NVOICES+1000]) () ||
		  STATE
		    where VAL:Num = val (tk_get "pan");
		= STATE;

rev_cb STATE	= put REV VAL ||
		  ifelse (get PLAYING) (n_set 2090 ("rev",VAL)) () || STATE
		    where VAL:Num = val (tk_get "rev");
		= STATE;

vol_cb STATE	= put GAIN VAL ||
		  ifelse (get PLAYING) (n_set 2098 ("gain",VAL)) () || STATE
		    where VAL:Num = val (tk_get "vol");
		= STATE;

rec_cb STATE	= put REC VAL ||
		  ifelse (get PLAYING) (n_set 2099 ("gain",VAL)) () || STATE
		    where VAL:Num = val (tk_get "rec");
		= STATE;

bufsz_cb STATE	= put BUFSZ SZ ||
		  ifelse (get PLAYING) (midictrl continue) () ||
		  STATE if SZ<>get BUFSZ
		    where SZ:Int = val (tk_get "bufsz");
		= STATE otherwise;
